#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Initialize DART
# Adrien Crovato

def initDart(cfg, scenario='aerodynamic', task='analysis'):
    """Initlialize DART for various API

    Parameters
    ----------
    cfg : dict
        Dictionary of parameters to configure DART
    scenario : string, optional
        Type of scenario (available: aerodynamic, aerostructural) (default: aerodynamic)
    task : string, optional
        Type of task (available: analysis, optimization) (default: analysis)

    Returns
    -------
    _dim : int
        Dimension of the problem
    _qinf : float
        Freestream dynamic pressure
    _msh : tbox.MshData object
        Mesh data structure
    _wrtr : tbox.MshExport object
        Mesh writer
    _mrf : tbox.MshDeform object
        Mesh morpher
    _pbl : dart.Probem object
        Problem definition
    _bnd : Dart.Body object
        Body of interest
    _sol : dart.Newton object
        Direct (Newton) solver
    _adj : dart.Adjoint object
        Adjoint solver
    """
    # Imports
    import dart
    import tbox
    import tbox.gmsh as gmsh
    import math

    # Basic checks and config
    # dimension
    if cfg['Dim'] != 2 and cfg['Dim'] != 3:
        raise RuntimeError('Problem dimension should be 2 or 3, but ' + cfg['Dim'] + ' was given!\n')
    else:
        _dim = cfg['Dim']
    # aoa and aos
    if 'AoA' in cfg:
        alpha = cfg['AoA'] * math.pi / 180
    else:
        alpha = 0
    if _dim == 2:
        if 'AoS' in cfg and cfg['AoS'] != 0:
            raise RuntimeError('Angle of sideslip (AoS) should be zero for 2D problems!\n')
        else:
            beta = 0
        if 'Symmetry' in cfg:
            raise RuntimeError('Symmetry boundary condition cannot be used for 2D problems!\n')
    else:
        if 'AoS' in cfg:
            if cfg['AoS'] != 0 and 'Symmetry' in cfg:
                raise RuntimeError('Symmetry boundary condition cannot be used with nonzero angle of sideslip (AoS)!\n')
            beta = cfg['AoS'] * math.pi / 180
        else:
            beta = 0
    # save format
    if cfg['Format'] == 'vtk':
        try:
            import tboxVtk
            Writer = tboxVtk.VtkExport
            print('VTK libraries found! Results will be saved in VTK format.\n')
        except:
            Writer = tbox.GmshExport
            print('VTK libraries not found! Results will be saved in gmsh format.\n')
    else:
        Writer = tbox.GmshExport
        print('Results will be saved in gmsh format.\n')
    # number of threads
    if 'Threads' in cfg:
        nthrd = cfg['Threads']
    else:
        nthrd = 1
    # verbosity
    if 'Verb' in cfg:
        verb = cfg['Verb']
    else:
        verb = 1
    # scenario and task type
    if scenario != 'aerodynamic' and scenario != 'aerostructural':
        raise RuntimeError('Scenario should be aerodynamic or aerostructural, but "' + scenario + '" was given!\n')
    if task != 'analysis' and task != 'optimization':
        raise RuntimeError('Task should be analysis or optimization, but "' + task + '" was given!\n')
    # dynamic pressure
    if scenario == 'aerostructural':
        _qinf = cfg['Q_inf']
    else:
        _qinf = None

    # Mesh creation
    _msh = gmsh.MeshLoader(cfg['File']).execute(**cfg['Pars'])
    # add the wake
    if _dim == 2:
        mshCrck = tbox.MshCrack(_msh, _dim)
        mshCrck.setCrack(cfg['Wake'])
        mshCrck.addBoundary(cfg['Fluid'])
        mshCrck.addBoundary(cfg['Farfield'][-1])
        mshCrck.addBoundary(cfg['Wing'])
        mshCrck.run()
    else:
        for i in range(len(cfg['Wakes'])):
            mshCrck = tbox.MshCrack(_msh, _dim)
            mshCrck.setCrack(cfg['Wakes'][i])
            mshCrck.addBoundary(cfg['Fluid'])
            mshCrck.addBoundary(cfg['Farfield'][-1])
            mshCrck.addBoundary(cfg['Wings'][i])
            if 'Fuselage' in cfg:
                mshCrck.addBoundary(cfg['Fuselage'])
            if 'Symmetry' in cfg:
                mshCrck.addBoundary(cfg['Symmetry'])
            if 'WakeTips' in cfg:
                mshCrck.setExcluded(cfg['WakeTips'][i]) # 3D computations (not excluded for 2.5D)
            mshCrck.run()
    # save the updated mesh in native (gmsh) format and then set the writer to the requested format
    tbox.GmshExport(_msh).save(_msh.name)
    del mshCrck
    _wrtr = Writer(_msh)
    print('\n')

    # Mesh morpher creation
    if scenario == 'aerostructural' or task == 'optimization':
        _mrf = tbox.MshDeform(_msh, tbox.Gmres(1, 1e-6, 30, 1e-8), _dim, nthrds=nthrd, vrb=verb)
        _mrf.setField(cfg['Fluid'])
        for tag in cfg['Farfield']:
            _mrf.addFixed(tag)
        if _dim == 2:
            _mrf.addMoving(cfg['Wing'])
            _mrf.addInternal(cfg['Wake'], cfg['Wake']+'_')
        else:
            if 'Fuselage' in cfg:
                _mrf.addFixed(cfg['Fuselage'])
            if 'Symmetry' in cfg: 
                _mrf.setSymmetry(cfg['Symmetry'], 1)
            for i in range(len(cfg['Wings'])):
                if i == 0:
                    _mrf.addMoving(cfg['Wings'][i]) # TODO body of interest (FSI/OPT) is always first given body
                else:
                    _mrf.addFixed(cfg['Wings'][i])
                _mrf.addInternal(cfg['Wakes'][i], cfg['Wakes'][i]+'_')
        _mrf.initialize()
    else:
        _mrf = None

    # Problem creation
    _pbl = dart.Problem(_msh, _dim, alpha, beta, cfg['M_inf'], cfg['S_ref'], cfg['c_ref'], cfg['x_ref'], cfg['y_ref'], cfg['z_ref'])
    # add the field
    _pbl.set(dart.Fluid(_msh, cfg['Fluid'], cfg['M_inf'], _dim, alpha, beta))
    # add the initial condition
    _pbl.set(dart.Initial(_msh, cfg['Fluid'], _dim, alpha, beta))
    # add farfield boundary conditions (Neumann only, a DOF will be pinned automatically)
    for i in range (len(cfg['Farfield'])):
        _pbl.add(dart.Freestream(_msh, cfg['Farfield'][i], _dim, alpha, beta))
    # add solid boundaries
    if _dim == 2:
        bnd = dart.Body(_msh, [cfg['Wing'], cfg['Fluid']])
        _bnd = bnd
        _pbl.add(bnd)
    else:
        _bnd = None
        for bd in cfg['Wings']:
            bnd = dart.Body(_msh, [bd, cfg['Fluid']])
            if _bnd is None:
                _bnd = bnd # TODO body of interest (FSI/OPT) is always first given body
            _pbl.add(bnd)
        if 'Fuselage' in cfg:
            bnd = dart.Body(_msh, [cfg['Fuselage'], cfg['Fluid']])
            _pbl.add(bnd)
    # add Wake/Kutta boundary conditions
    # TODO refactor Wake "exclusion" for 3D?
    if _dim == 2:
        _pbl.add(dart.Wake(_msh, [cfg['Wake'], cfg['Wake']+'_', cfg['Fluid']]))
        _pbl.add(dart.Kutta(_msh, [cfg['Te'], cfg['Wake']+'_', cfg['Wing'], cfg['Fluid']]))
    else:
        for i in range(len(cfg['Wakes'])):
            if 'WakeExs' in cfg:
                _pbl.add(dart.Wake(_msh, [cfg['Wakes'][i], cfg['Wakes'][i]+'_', cfg['Fluid'], cfg['WakeExs'][i]])) # 3D + fuselage
            elif 'WakeTips' in cfg:
                _pbl.add(dart.Wake(_msh, [cfg['Wakes'][i], cfg['Wakes'][i]+'_', cfg['Fluid'], cfg['WakeTips'][i]])) # 3D
            else:
                _pbl.add(dart.Wake(_msh, [cfg['Wakes'][i], cfg['Wakes'][i]+'_', cfg['Fluid']])) # 2.5D
            _pbl.add(dart.Kutta(_msh, [cfg['Tes'][i], cfg['Wakes'][i]+'_', cfg['Wings'][i], cfg['Fluid']]))

    # Direct (forward) solver creation
    # NB: more solvers/options are available but we restrict the user's choice to the most efficient ones
    # initialize the linear (inner) solver
    if cfg['LSolver'] == 'PARDISO':
        linsol = tbox.Pardiso()
    elif cfg['LSolver'] == 'MUMPS':
        linsol = tbox.Mumps()
    elif cfg['LSolver'] == 'GMRES':
        gfil = cfg['G_fill'] if 'G_fill' in cfg else 2
        grst = cfg['G_restart'] if 'G_restart' in cfg else 50
        gtol = cfg['G_tol'] if 'G_tol' in cfg else 1e-5
        linsol = tbox.Gmres(gfil, 1e-6, grst, gtol)
    else:
        raise RuntimeError('Available linear solvers: PARDISO, MUMPS or GMRES, but ' + cfg['LSolver'] + ' was given!\n')
    # initialize the nonlinear (outer) solver
    _sol = dart.Newton(_pbl, linsol, rTol=cfg['Rel_tol'], aTol=cfg['Abs_tol'], mIt=cfg['Max_it'], nthrds=nthrd, vrb=verb)

    # Adjoint (reverse) solver creation
    if task == 'optimization':
        _adj = dart.Adjoint(_sol, _mrf)
    else:
        _adj = None

    # Return
    return _dim, _qinf, _msh, _wrtr, _mrf, _pbl, _bnd, _sol, _adj
